package com.sun.xml.internal.ws.server;

import com.sun.xml.internal.ws.api.pipe.Tube;
import javax.xml.ws.WebServiceException;
import com.sun.xml.internal.ws.resources.ServerMessages;
import com.sun.xml.internal.ws.api.pipe.helper.AbstractTubeImpl;
import com.sun.xml.internal.ws.api.pipe.TubeCloner;
import com.sun.istack.internal.NotNull;
import com.sun.xml.internal.ws.api.server.WSWebServiceContext;
import com.sun.istack.internal.Nullable;
import javax.xml.ws.WebServiceContext;
import com.sun.xml.internal.ws.api.server.AsyncProviderCallback;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import com.sun.xml.internal.ws.api.message.Packet;
import com.sun.xml.internal.ws.api.server.WSEndpoint;
import com.sun.xml.internal.ws.api.server.Invoker;

public abstract class InvokerTube<T> extends com.sun.xml.internal.ws.server.sei.InvokerTube<Invoker> implements EndpointAwareTube
{
    private WSEndpoint endpoint;
    private static final ThreadLocal<Packet> packets;
    private final Invoker wrapper;
    
    protected InvokerTube(Invoker invoker) {
        super(invoker);
        this.wrapper = new Invoker() {
            public Object invoke(Packet p, Method m, Object... args) throws InvocationTargetException, IllegalAccessException {
                Packet old = this.set(p);
                try {
                    return ((Invoker)InvokerTube.this.invoker).invoke(p, m, args);
                }
                finally {
                    this.set(old);
                }
            }
            
            public <T> T invokeProvider(Packet p, T arg) throws IllegalAccessException, InvocationTargetException {
                Packet old = this.set(p);
                try {
                    return ((Invoker)InvokerTube.this.invoker).invokeProvider(p, arg);
                }
                finally {
                    this.set(old);
                }
            }
            
            public <T> void invokeAsyncProvider(Packet p, T arg, AsyncProviderCallback cbak, WebServiceContext ctxt) throws IllegalAccessException, InvocationTargetException {
                Packet old = this.set(p);
                try {
                    ((Invoker)InvokerTube.this.invoker).invokeAsyncProvider(p, arg, cbak, ctxt);
                }
                finally {
                    this.set(old);
                }
            }
            
            private Packet set(Packet p) {
                Packet old = (Packet)InvokerTube.packets.get();
                InvokerTube.packets.set(p);
                return old;
            }
        };
    }
    
    public void setEndpoint(final WSEndpoint endpoint) {
        this.endpoint = endpoint;
        WSWebServiceContext webServiceContext = new AbstractWebServiceContext(endpoint) {
            @Nullable
            public Packet getRequestPacket() {
                Packet p = (Packet)InvokerTube.packets.get();
                return p;
            }
        };
        ((Invoker)this.invoker).start(webServiceContext, endpoint);
    }
    
    protected WSEndpoint getEndpoint() {
        return this.endpoint;
    }
    
    @NotNull
    public final Invoker getInvoker(Packet request) {
        return this.wrapper;
    }
    
    public final AbstractTubeImpl copy(TubeCloner cloner) {
        cloner.add(this, this);
        return this;
    }
    
    public void preDestroy() {
        ((Invoker)this.invoker).dispose();
    }
    
    @NotNull
    public static Packet getCurrentPacket() {
        Packet packet = (Packet)InvokerTube.packets.get();
        if (packet == null) {
            throw new WebServiceException(ServerMessages.NO_CURRENT_PACKET());
        }
        return packet;
    }
    
    static {
        packets = new ThreadLocal();
    }
}
